import pkg_resources

from .question_answering import BertQuestionAnsweringModel

__version__ = (
    pkg_resources.resource_string("adaptnlp", "VERSION.txt").decode("UTF-8").strip()
)

__all__ = [
    "__version__",
    "BertQuestionAnsweringModel",
]



import logging
from typing import Tuple, List, Union, Dict
from collections import OrderedDict, defaultdict
from tqdm import tqdm

import torch
from torch.utils.data import DataLoader
from transformers import (
    AutoTokenizer,
    AutoModelForQuestionAnswering,
    XLNetForQuestionAnswering,
    XLMForQuestionAnswering,
    CamembertForQuestionAnswering,
    DistilBertForQuestionAnswering,
    RobertaForQuestionAnswering,
    PreTrainedModel,
    PreTrainedTokenizer,
    SquadExample,
    squad_convert_examples_to_features,
)
from transformers.data.processors.squad import SquadResult

from adaptnlp.transformers import BertQuestionAnsweringModel
from adaptnlp.model import AdaptiveModel
from adaptnlp.transformers.squad_metrics import (
    compute_predictions_log_probs,
    compute_predictions_logits,
)

logger = logging.getLogger(__name__)


class TransformersQuestionAnswering(AdaptiveModel):
    """Adaptive Model for Transformers Question Answering Model

    **Parameters**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers *
    * **model** - A transformer Question Answering model

    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        self.model = model

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

        self.xmodel_instances = (XLNetForQuestionAnswering, XLMForQuestionAnswering)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this model

        * **model_name_or_path** - A key string of one of Transformer's pre-trained Question Answering (SQUAD) models
        """
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForQuestionAnswering.from_pretrained(model_name_or_path)
        qa_model = cls(tokenizer, model)
        return qa_model

    def predict(
        self,
        query: Union[List[str], str],
        context: Union[List[str], str],
        n_best_size: int = 5,
        mini_batch_size: int = 32,
        max_answer_length: int = 10,
        do_lower_case: bool = False,
        version_2_with_negative: bool = False,
        verbose_logging: bool = False,
        null_score_diff_threshold: float = 0.0,
        max_seq_length: int = 512,
        doc_stride: int = 128,
        max_query_length: int = 64,
        **kwargs,
    ) -> Tuple[Tuple[str, List[OrderedDict]], Tuple[OrderedDict, OrderedDict]]:
        """Predict method for running inference using the pre-trained question answering model

        * **query** - String or list of strings that specify the ordered questions corresponding to `context`
        * **context** - String or list of strings that specify the ordered contexts corresponding to `query`
        * **n_best_size** - Number of top n results you want
        * **mini_batch_size** - Mini batch size
        * **max_answer_length** - Maximum token length for answers that are returned
        * **do_lower_case** - Set as `True` if using uncased QA models
        * **version_2_with_negative** - Set as True if using QA model with SQUAD2.0
        * **verbose_logging** - Set True if you want prediction verbose loggings
        * **null_score_diff_threshold** - Threshold for predicting null(no answer) in Squad 2.0 Model.  Default is 0.0.  Raise this if you want fewer null answers
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        * **doc_stride** - Number of token strides to take when splitting up conext into chunks of size `max_seq_length`
        * **max_query_length** - Maximum token length for queries
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers model (mostly for saving evaluations)
        """
        # Make string input consistent as list
        if isinstance(query, str):
            query = [query]
            context = [context]
        assert len(query) == len(context)

        examples = self._mini_squad_processor(query=query, context=context)
        features, dataset = squad_convert_examples_to_features(
            examples,
            self.tokenizer,
            max_seq_length=max_seq_length,
            doc_stride=doc_stride,
            max_query_length=max_query_length,
            is_training=False,
            return_dataset="pt",
            threads=1,
        )
        all_results = []

        with torch.no_grad():

            dataloader = DataLoader(dataset, batch_size=mini_batch_size)

            for batch in tqdm(dataloader, desc="Predicting answer"):
                self.model.eval()
                batch = tuple(t.to(self.device) for t in batch)

                inputs = {
                    "input_ids": batch[0],
                    "attention_mask": batch[1],
                    "token_type_ids": batch[2],
                }
                example_indices = batch[3]

                if isinstance(
                    self.model,
                    (
                        XLMForQuestionAnswering,
                        RobertaForQuestionAnswering,
                        DistilBertForQuestionAnswering,
                        CamembertForQuestionAnswering,
                    ),
                ):
                    del inputs["token_type_ids"]

                # XLNet and XLM use more arguments for their predictions
                if isinstance(self.model, self.xmodel_instances):
                    inputs.update({"cls_index": batch[4], "p_mask": batch[5]})
                    # for lang_id-sensitive xlm models
                    if hasattr(self.model, "config") and hasattr(
                        self.model.config, "lang2id"
                    ):
                        # Set language id as 0 for now
                        inputs.update(
                            {
                                "langs": (
                                    torch.ones(batch[0].shape, dtype=torch.int64) * 0
                                ).to(self.device)
                            }
                        )

                outputs = self.model(**inputs)

                # Iterate through and produce `SquadResults
                for i, example_index in enumerate(example_indices):
                    eval_feature = features[example_index.item()]
                    unique_id = int(eval_feature.unique_id)

                    output = [self.to_list(output[i]) for output in outputs]

                    if isinstance(self.model, self.xmodel_instances):
                        # Some models like the ones in `self.xmodel_instances` use 5 arguments for their predictions
                        start_logits = output[0]
                        start_top_index = output[1]
                        end_logits = output[2]
                        end_top_index = output[3]
                        cls_logits = output[4]

                        result = SquadResult(
                            unique_id,
                            start_logits,
                            end_logits,
                            start_top_index=start_top_index,
                            end_top_index=end_top_index,
                            cls_logits=cls_logits,
                        )

                    else:
                        start_logits, end_logits = output
                        result = SquadResult(unique_id, start_logits, end_logits)
                    all_results.append(result)

            if isinstance(self.model, self.xmodel_instances):
                start_n_top = (
                    self.model.config.start_n_top
                    if hasattr(self.model, "config")
                    else self.model.module.config.start_n_top
                )
                end_n_top = (
                    self.model.config.end_n_top
                    if hasattr(self.model, "config")
                    else self.model.module.config.end_n_top
                )

                answers, n_best = compute_predictions_log_probs(
                    examples,
                    features,
                    all_results,
                    n_best_size,
                    max_answer_length,
                    start_n_top,
                    end_n_top,
                    version_2_with_negative,
                    self.tokenizer,
                    verbose_logging,
                    **kwargs,
                )

            else:
                answers, n_best = compute_predictions_logits(
                    examples,
                    features,
                    all_results,
                    n_best_size,
                    max_answer_length,
                    do_lower_case,
                    verbose_logging,
                    version_2_with_negative,
                    null_score_diff_threshold,
                    self.tokenizer,
                    **kwargs,
                )

        return answers, n_best

    def to_list(self, tensor: torch.Tensor):
        return tensor.detach().cpu().tolist()

    def _mini_squad_processor(
        self, query: List[str], context: List[str]
    ) -> List[SquadExample]:
        """Squad data processor to create `SquadExamples`

        * **query** - List of query strings, must be same length as `context`
        * **context** - List of context strings, must be same length as `query`

        """
        assert len(query) == len(context)
        examples = []
        title = "qa"
        is_impossible = False
        answer_text = None
        start_position_character = None
        answers = ["answer"]
        for idx, (q, c) in enumerate(zip(query, context)):
            example = SquadExample(
                qas_id=str(idx),
                question_text=q,
                context_text=c,
                answer_text=answer_text,
                start_position_character=start_position_character,
                title=title,
                is_impossible=is_impossible,
                answers=answers,
            )
            examples.append(example)
        return examples

    def train(
        self,
    ):
        raise NotImplementedError

    def evaluate(
        self,
    ):
        raise NotImplementedError


class EasyQuestionAnswering:
    """Question Answering Module

    Usage:

    ```python
    >>> qa = adaptnlp.EasyQuestionAnswering()
    >>> qa.predict_qa(query="What is life?", context="Life is NLP.", n_best_size=5, mini_batch_size=1)
    ```
    """

    def __init__(self):
        self.models: Dict[AdaptiveModel] = defaultdict(bool)

        # Soon to deprecated
        self.bert_qa = None

    # Dynamic Loaders
    def _load_bert_qa(self) -> None:
        self.bert_qa = BertQuestionAnsweringModel()

    # Soon to be deprecated
    def predict_bert_qa(
        self, query: str, context: str, n_best_size: int = 20
    ) -> Tuple[str, List[OrderedDict]]:
        """Predicts top_n answer spans of query in regards to context


        * **query** - The question
        * **context** - The context of which the question is asking
        * **n_best_size** - The top n answers returned

        **return** - Either a list of string answers or a dict of the results
        """

        self._load_bert_qa() if not self.bert_qa else None
        return self.bert_qa.predict(
            query=query, context=context, n_best_size=n_best_size
        )

    def predict_qa(
        self,
        query: Union[List[str], str],
        context: Union[List[str], str],
        n_best_size: int = 5,
        mini_batch_size: int = 32,
        model_name_or_path: str = "bert-large-uncased-whole-word-masking-finetuned-squad",
        **kwargs,
    ) -> Tuple[Tuple[str, List[OrderedDict]], Tuple[OrderedDict, OrderedDict]]:
        """Predicts top_n answer spans of query in regards to context

        * **query** - String or list of strings that specify the ordered questions corresponding to `context`
        * **context** - String or list of strings that specify the ordered contexts corresponding to `query`
        * **n_best_size** - The top n answers returned
        * **mini_batch_size** - Mini batch size for inference
        * **model_name_or_path** - Path to QA model or name of QA model at huggingface.co/models
        * **kwargs**(Optional) - Keyword arguments for `AdaptiveModel`s like `TransformersQuestionAnswering`

        **return** - Either a list of string answers or a dict of the results
        """
        try:
            if not self.models[model_name_or_path]:
                self.models[model_name_or_path] = TransformersQuestionAnswering.load(
                    model_name_or_path
                )
        except OSError:
            logger.info(
                f"{model_name_or_path} not a valid Transformers pre-trained QA model...check path or huggingface.co/models"
            )
            raise ValueError(
                f"{model_name_or_path} is not a valid path or model name from huggingface.co/models"
            )
            return OrderedDict(), [OrderedDict()]

        model = self.models[model_name_or_path]

        # If query and context is just one instance, get rid of nested OrderedDict
        if isinstance(query, str):
            top_answer, top_n_answers = model.predict(
                query=query,
                context=context,
                n_best_size=n_best_size,
                mini_batch_size=mini_batch_size,
                **kwargs,
            )
            return top_answer["0"], top_n_answers["0"]

        return model.predict(
            query=query,
            context=context,
            n_best_size=n_best_size,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )



# Contains code used/modified by AdaptNLP author from transformers
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Union, Tuple
import collections
from collections import OrderedDict

import torch
from torch.utils.data import DataLoader, SequentialSampler, TensorDataset

from transformers import (
    BertConfig,
    BertForQuestionAnswering,
    BertTokenizer,
    XLMConfig,
    XLMForQuestionAnswering,
    XLMTokenizer,
    XLNetConfig,
    XLNetForQuestionAnswering,
    XLNetTokenizer,
)

from adaptnlp.transformers.utils_squad import (
    SquadExample,
    InputFeatures,
    convert_examples_to_features,
    RawResult,
    RawResultExtended,
    get_final_text,
    _get_best_indexes,
    _compute_softmax,
)


class QuestionAnsweringModel(ABC):
    @abstractmethod
    def __init__(self):
        super().__init__()
        self.config
        self.tokenizer
        self.model

    @abstractmethod
    def _load(self):
        raise NotImplementedError

    @abstractmethod
    def predict(self, query, context, top_n, as_dict):
        raise NotImplementedError


# TODO To be deprecated in the near future for a better module design
class BertQuestionAnsweringModel(QuestionAnsweringModel):
    def __init__(self):
        self.config = BertConfig
        self.tokenizer = BertTokenizer
        self.model = BertForQuestionAnswering
        self.model_names = list(self.config.pretrained_config_archive_map.keys())

        # Post Load
        self.pretrained_config = None
        self.pretrained_tokenizer = None
        self.pretrained_model = None

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def _to_list(self, tensor: torch.Tensor) -> List[float]:
        return tensor.detach().cpu().tolist()

    def _load(self) -> None:
        print("Loading Pretrained Bert Question Answering Model...")
        model_name = "bert-large-uncased-whole-word-masking-finetuned-squad"
        self.pretrained_config = self.config.from_pretrained(
            "bert-large-uncased-whole-word-masking-finetuned-squad"
        )
        if "uncased" in model_name:
            tokenizer = self.tokenizer.from_pretrained(
                "bert-large-uncased", do_lower_case=True
            )
        else:
            tokenizer = self.tokenizer.from_pretrained(
                "bert-large-cased", do_lower_case=False
            )
        self.pretrained_tokenizer = tokenizer

        model = self.model.from_pretrained(
            model_name,
            from_tf=bool(".ckpt" in model_name),
            config=self.pretrained_config,
        )
        self.pretrained_model = model
        self.pretrained_model.to(self.device)

    def _load_one_query(
        self, query: str, context: str, output_examples=True
    ) -> Union[TensorDataset, List[SquadExample], List[InputFeatures]]:
        # Create doc_tokens for SquadExample with one query and context

        def is_whitespace(c):
            if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
                return True
            return False

        # Create doc_tokens
        doc_tokens = []
        prev_is_whitespace = True
        for c in context:
            if is_whitespace(c):
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    doc_tokens.append(c)
                else:
                    doc_tokens[-1] += c
                prev_is_whitespace = False

        # Create SquadExample
        examples = []
        example = SquadExample(
            qas_id=None,
            question_text=query,
            doc_tokens=doc_tokens,
            orig_answer_text=None,
            start_position=None,
            end_position=None,
            is_impossible=False,
        )
        examples.append(example)

        # Convert to features
        features = convert_examples_to_features(
            examples=examples,
            tokenizer=self.pretrained_tokenizer,
            max_seq_length=384,
            doc_stride=128,
            max_query_length=64,
            is_training=False,
        )

        # Convert to Tensors and build dataset
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_input_mask = torch.tensor(
            [f.input_mask for f in features], dtype=torch.long
        )
        all_segment_ids = torch.tensor(
            [f.segment_ids for f in features], dtype=torch.long
        )
        all_cls_index = torch.tensor([f.cls_index for f in features], dtype=torch.long)
        all_p_mask = torch.tensor([f.p_mask for f in features], dtype=torch.float)

        all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)
        dataset = TensorDataset(
            all_input_ids,
            all_input_mask,
            all_segment_ids,
            all_example_index,
            all_cls_index,
            all_p_mask,
        )

        if output_examples:
            return dataset, examples, features
        return dataset

    def _produce_concrete_predictions(
        self,
        all_examples,
        all_features,
        all_results,
        n_best_size=10,
        max_answer_length=30,
        do_lower_case=True,
        verbose_logging=False,
        version_2_with_negative=True,
        null_score_diff_threshold=0.0,
    ):

        example_index_to_features = collections.defaultdict(list)
        for feature in all_features:
            example_index_to_features[feature.example_index].append(feature)

        unique_id_to_result = {}
        for result in all_results:
            unique_id_to_result[result.unique_id] = result

        _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "PrelimPrediction",
            ["feature_index", "start_index", "end_index", "start_logit", "end_logit"],
        )

        all_predictions = collections.OrderedDict()
        all_nbest_json = collections.OrderedDict()
        scores_diff_json = collections.OrderedDict()

        for (example_index, example) in enumerate(all_examples):
            features = example_index_to_features[example_index]

            prelim_predictions = []
            # keep track of the minimum score of null start+end of position 0
            score_null = 1000000  # large and positive
            min_null_feature_index = 0  # the paragraph slice with min null score
            null_start_logit = 0  # the start logit at the slice with min null score
            null_end_logit = 0  # the end logit at the slice with min null score
            for (feature_index, feature) in enumerate(features):
                result = unique_id_to_result[feature.unique_id]
                start_indexes = _get_best_indexes(result.start_logits, n_best_size)
                end_indexes = _get_best_indexes(result.end_logits, n_best_size)
                # if we could have irrelevant answers, get the min score of irrelevant
                if version_2_with_negative:
                    feature_null_score = result.start_logits[0] + result.end_logits[0]
                    if feature_null_score < score_null:
                        score_null = feature_null_score
                        min_null_feature_index = feature_index
                        null_start_logit = result.start_logits[0]
                        null_end_logit = result.end_logits[0]
                for start_index in start_indexes:
                    for end_index in end_indexes:
                        # We could hypothetically create invalid predictions, e.g., predict
                        # that the start of the span is in the question. We throw out all
                        # invalid predictions.
                        if start_index >= len(feature.tokens):
                            continue
                        if end_index >= len(feature.tokens):
                            continue
                        if start_index not in feature.token_to_orig_map:
                            continue
                        if end_index not in feature.token_to_orig_map:
                            continue
                        if not feature.token_is_max_context.get(start_index, False):
                            continue
                        if end_index < start_index:
                            continue
                        length = end_index - start_index + 1
                        if length > max_answer_length:
                            continue
                        prelim_predictions.append(
                            _PrelimPrediction(
                                feature_index=feature_index,
                                start_index=start_index,
                                end_index=end_index,
                                start_logit=result.start_logits[start_index],
                                end_logit=result.end_logits[end_index],
                            )
                        )
            if version_2_with_negative:
                prelim_predictions.append(
                    _PrelimPrediction(
                        feature_index=min_null_feature_index,
                        start_index=0,
                        end_index=0,
                        start_logit=null_start_logit,
                        end_logit=null_end_logit,
                    )
                )
            prelim_predictions = sorted(
                prelim_predictions,
                key=lambda x: (x.start_logit + x.end_logit),
                reverse=True,
            )

            _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
                "NbestPrediction",
                ["text", "start_index", "end_index", "start_logit", "end_logit"],
            )  # ### start_end_index

            seen_predictions = {}
            nbest = []
            for pred in prelim_predictions:
                if len(nbest) >= n_best_size:
                    break
                feature = features[pred.feature_index]
                orig_doc_start = 0
                orig_doc_end = 0
                if pred.start_index > 0:  # this is a non-null prediction
                    tok_tokens = feature.tokens[pred.start_index : (pred.end_index + 1)]
                    orig_doc_start = feature.token_to_orig_map[pred.start_index]
                    orig_doc_end = feature.token_to_orig_map[pred.end_index]
                    orig_tokens = example.doc_tokens[
                        orig_doc_start : (orig_doc_end + 1)
                    ]
                    tok_text = " ".join(tok_tokens)

                    # De-tokenize WordPieces that have been split off.
                    tok_text = tok_text.replace(" ##", "")
                    tok_text = tok_text.replace("##", "")

                    # Clean whitespace
                    tok_text = tok_text.strip()
                    tok_text = " ".join(tok_text.split())
                    orig_text = " ".join(orig_tokens)

                    final_text = get_final_text(
                        tok_text, orig_text, do_lower_case, verbose_logging
                    )
                    if final_text in seen_predictions:
                        continue

                    seen_predictions[final_text] = True
                else:
                    final_text = ""
                    seen_predictions[final_text] = True

                nbest.append(
                    _NbestPrediction(
                        text=final_text,
                        start_logit=pred.start_logit,
                        end_logit=pred.end_logit,
                        start_index=orig_doc_start,
                        end_index=orig_doc_end,
                    )
                )  # ### start_end_index...Make span indices inclusive
            # if we didn't include the empty option in the n-best, include it
            if version_2_with_negative:
                if "" not in seen_predictions:
                    nbest.append(
                        _NbestPrediction(
                            text="",
                            start_logit=null_start_logit,
                            end_logit=null_end_logit,
                            start_index=0,
                            end_index=0,
                        )
                    )  # ### start_end_index should this be pred.<index>

                # In very rare edge cases we could only have single null prediction.
                # So we just create a nonce prediction in this case to avoid failure.
                if len(nbest) == 1:
                    nbest.insert(
                        0,
                        _NbestPrediction(
                            text="empty",
                            start_logit=0.0,
                            end_logit=0.0,
                            start_index=0.0,
                            end_index=0.0,
                        ),
                    )  # ### start_end_index

            # In very rare edge cases we could have no valid predictions. So we
            # just create a nonce prediction in this case to avoid failure.
            if not nbest:
                nbest.append(
                    _NbestPrediction(
                        text="empty",
                        start_logit=0.0,
                        end_logit=0.0,
                        start_index=0.0,
                        end_index=0.0,
                    )
                )  # ### start_end_index

            assert len(nbest) >= 1

            total_scores = []
            best_non_null_entry = None
            for entry in nbest:
                total_scores.append(entry.start_logit + entry.end_logit)
                if not best_non_null_entry:
                    if entry.text:
                        best_non_null_entry = entry

            probs = _compute_softmax(total_scores)

            nbest_json = []
            for (i, entry) in enumerate(nbest):
                output = collections.OrderedDict()
                output["text"] = entry.text
                output["probability"] = probs[i]
                output["start_logit"] = entry.start_logit
                output["end_logit"] = entry.end_logit
                output[
                    "start_index"
                ] = (
                    entry.start_index
                )  # ### start_end_index MAGIC NUMBERS for adjustment :/
                output["end_index"] = entry.end_index
                nbest_json.append(output)

            assert len(nbest_json) >= 1

            if not version_2_with_negative:
                all_predictions[example.qas_id] = nbest_json[0]["text"]
            else:
                # predict "" iff the null score - the score of best non-null > threshold
                score_diff = (
                    score_null
                    - best_non_null_entry.start_logit
                    - (best_non_null_entry.end_logit)
                )
                scores_diff_json[example.qas_id] = score_diff
                if score_diff > null_score_diff_threshold:
                    all_predictions[example.qas_id] = ""
                else:
                    all_predictions[example.qas_id] = best_non_null_entry.text
            all_nbest_json[example.qas_id] = nbest_json

            # All ids set as None so get rid of None Key
            all_predictions = all_predictions[None]
            all_nbest_json = all_nbest_json[None]

        return all_predictions, all_nbest_json

    def predict(
        self, query: str, context: str, n_best_size: int = 20
    ) -> Tuple[str, List[OrderedDict]]:
        """Predicts top_n answer spans of query in regards to context

        Args:
            query: The question
            context: The context of which the question is asking
            top_n: The top n answers returned

        Returns:
            Either a list of string answers or a dict of the results
        """
        self._load() if not self.pretrained_model or not self.pretrained_tokenizer else None

        # Load and Evaluate Context Queries
        dataset, examples, features = self._load_one_query(query, context)
        eval_sampler = SequentialSampler(dataset)
        eval_dataloader = DataLoader(dataset, sampler=eval_sampler, batch_size=5)
        all_results = []
        for batch in eval_dataloader:
            self.pretrained_model.eval()
            batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                # BERT XLM XLNET DIFFERENCE
                inputs = {
                    "input_ids": batch[0],
                    "attention_mask": batch[1],
                    "token_type_ids": batch[2],
                }
                example_indices = batch[3]
                outputs = self.pretrained_model(**inputs)

            for i, example_index in enumerate(example_indices):
                eval_feature = features[example_index.item()]
                unique_id = int(eval_feature.unique_id)
                # BERT XLM XLNET DIFFERENCE
                result = RawResult(
                    unique_id=unique_id,
                    start_logits=self._to_list(outputs[0][i]),
                    end_logits=self._to_list(outputs[1][i]),
                )
                all_results.append(result)

        # Obtain Concrete Predictions
        all_predictions, all_nbest_json = self._produce_concrete_predictions(
            examples, features, all_results, n_best_size=n_best_size
        )
        return all_predictions, all_nbest_json


class XLNetQuestionAnsweringModel(QuestionAnsweringModel):
    def __init__(self):
        self.config = XLNetConfig
        self.tokenizer = XLNetTokenizer
        self.model = XLNetForQuestionAnswering
        self.model_names = list(self.config.pretrained_config_archive_map.keys())

        # Post Load
        self.pretrained_config = None
        self.pretrained_tokenizer = None
        self.pretrained_model = None

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def _to_list(self, tensor: torch.Tensor) -> List[float]:
        return tensor.detach().cpu().tolist()

    def _load(self) -> None:
        print("Loading Pretrained XLNet Question Answering Model...")
        model_name = "xlnet-large-cased"
        self.pretrained_config = self.config.from_pretrained("xlnet-large-cased")
        tokenizer = self.tokenizer.from_pretrained(
            "xlnet-large-cased", do_lower_case=False
        )
        self.pretrained_tokenizer = tokenizer

        model = self.model.from_pretrained(
            model_name,
            from_tf=bool(".ckpt" in model_name),
            config=self.pretrained_config,
        )
        self.pretrained_model = model
        self.pretrained_model.to(self.device)

    def _load_one_query(
        self, query: str, context: str, output_examples=True
    ) -> Union[TensorDataset, List[SquadExample], List[InputFeatures]]:
        # Create doc_tokens for SquadExample with one query and context

        def is_whitespace(c):
            if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
                return True
            return False

        # Create doc_tokens
        doc_tokens = []
        prev_is_whitespace = True
        for c in context:
            if is_whitespace(c):
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    doc_tokens.append(c)
                else:
                    doc_tokens[-1] += c
                prev_is_whitespace = False

        # Create SquadExample
        examples = []
        example = SquadExample(
            qas_id=None,
            question_text=query,
            doc_tokens=doc_tokens,
            orig_answer_text=None,
            start_position=None,
            end_position=None,
            is_impossible=False,
        )
        examples.append(example)

        # Convert to features
        features = convert_examples_to_features(
            examples=examples,
            tokenizer=self.pretrained_tokenizer,
            max_seq_length=384,
            doc_stride=128,
            max_query_length=64,
            is_training=False,
        )

        # Convert to Tensors and build dataset
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_input_mask = torch.tensor(
            [f.input_mask for f in features], dtype=torch.long
        )
        all_segment_ids = torch.tensor(
            [f.segment_ids for f in features], dtype=torch.long
        )
        all_cls_index = torch.tensor([f.cls_index for f in features], dtype=torch.long)
        all_p_mask = torch.tensor([f.p_mask for f in features], dtype=torch.float)

        all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)
        dataset = TensorDataset(
            all_input_ids,
            all_input_mask,
            all_segment_ids,
            all_example_index,
            all_cls_index,
            all_p_mask,
        )

        if output_examples:
            return dataset, examples, features
        return dataset

    def _produce_concrete_predictions(
        self,
        all_examples,
        all_features,
        all_results,
        n_best_size=10,
        max_answer_length=30,
        verbose_logging=False,
    ):

        start_n_top = self.pretrained_model.config.start_n_top
        end_n_top = self.pretrained_model.config.end_n_top
        tokenizer = self.pretrained_tokenizer

        _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "PrelimPrediction",
            [
                "feature_index",
                "start_index",
                "end_index",
                "start_log_prob",
                "end_log_prob",
            ],
        )

        _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "NbestPrediction", ["text", "start_log_prob", "end_log_prob"]
        )

        example_index_to_features = collections.defaultdict(list)
        for feature in all_features:
            example_index_to_features[feature.example_index].append(feature)

        unique_id_to_result = {}
        for result in all_results:
            unique_id_to_result[result.unique_id] = result

        all_predictions = collections.OrderedDict()
        all_nbest_json = collections.OrderedDict()
        scores_diff_json = collections.OrderedDict()

        for (example_index, example) in enumerate(all_examples):
            features = example_index_to_features[example_index]

            prelim_predictions = []
            # keep track of the minimum score of null start+end of position 0
            score_null = 1000000  # large and positive

            for (feature_index, feature) in enumerate(features):
                result = unique_id_to_result[feature.unique_id]

                cur_null_score = result.cls_logits

                # if we could have irrelevant answers, get the min score of irrelevant
                score_null = min(score_null, cur_null_score)

                for i in range(start_n_top):
                    for j in range(end_n_top):
                        start_log_prob = result.start_top_log_probs[i]
                        start_index = result.start_top_index[i]

                        j_index = i * end_n_top + j

                        end_log_prob = result.end_top_log_probs[j_index]
                        end_index = result.end_top_index[j_index]

                        # We could hypothetically create invalid predictions, e.g., predict
                        # that the start of the span is in the question. We throw out all
                        # invalid predictions.
                        if start_index >= feature.paragraph_len - 1:
                            continue
                        if end_index >= feature.paragraph_len - 1:
                            continue

                        if not feature.token_is_max_context.get(start_index, False):
                            continue
                        if end_index < start_index:
                            continue
                        length = end_index - start_index + 1
                        if length > max_answer_length:
                            continue

                        prelim_predictions.append(
                            _PrelimPrediction(
                                feature_index=feature_index,
                                start_index=start_index,
                                end_index=end_index,
                                start_log_prob=start_log_prob,
                                end_log_prob=end_log_prob,
                            )
                        )

            prelim_predictions = sorted(
                prelim_predictions,
                key=lambda x: (x.start_log_prob + x.end_log_prob),
                reverse=True,
            )

            seen_predictions = {}
            nbest = []
            for pred in prelim_predictions:
                if len(nbest) >= n_best_size:
                    break
                feature = features[pred.feature_index]

                # XLNet un-tokenizer
                # Let's keep it simple for now and see if we need all this later.
                #
                # tok_start_to_orig_index = feature.tok_start_to_orig_index
                # tok_end_to_orig_index = feature.tok_end_to_orig_index
                # start_orig_pos = tok_start_to_orig_index[pred.start_index]
                # end_orig_pos = tok_end_to_orig_index[pred.end_index]
                # paragraph_text = example.paragraph_text
                # final_text = paragraph_text[start_orig_pos: end_orig_pos + 1].strip()

                # Previously used Bert untokenizer
                tok_tokens = feature.tokens[pred.start_index : (pred.end_index + 1)]
                orig_doc_start = feature.token_to_orig_map[pred.start_index]
                orig_doc_end = feature.token_to_orig_map[pred.end_index]
                orig_tokens = example.doc_tokens[orig_doc_start : (orig_doc_end + 1)]
                tok_text = tokenizer.convert_tokens_to_string(tok_tokens)

                # Clean whitespace
                tok_text = tok_text.strip()
                tok_text = " ".join(tok_text.split())
                orig_text = " ".join(orig_tokens)

                final_text = get_final_text(
                    tok_text, orig_text, tokenizer.do_lower_case, verbose_logging
                )

                if final_text in seen_predictions:
                    continue

                seen_predictions[final_text] = True

                nbest.append(
                    _NbestPrediction(
                        text=final_text,
                        start_log_prob=pred.start_log_prob,
                        end_log_prob=pred.end_log_prob,
                    )
                )

            # In very rare edge cases we could have no valid predictions. So we
            # just create a nonce prediction in this case to avoid failure.
            if not nbest:
                nbest.append(
                    _NbestPrediction(text="", start_log_prob=-1e6, end_log_prob=-1e6)
                )

            total_scores = []
            best_non_null_entry = None
            for entry in nbest:
                total_scores.append(entry.start_log_prob + entry.end_log_prob)
                if not best_non_null_entry:
                    best_non_null_entry = entry

            probs = _compute_softmax(total_scores)

            nbest_json = []
            for (i, entry) in enumerate(nbest):
                output = collections.OrderedDict()
                output["text"] = entry.text
                output["probability"] = probs[i]
                output["start_log_prob"] = entry.start_log_prob
                output["end_log_prob"] = entry.end_log_prob
                nbest_json.append(output)

            assert len(nbest_json) >= 1
            assert best_non_null_entry is not None

            score_diff = score_null
            scores_diff_json[example.qas_id] = score_diff
            # note(zhiliny): always predict best_non_null_entry
            # and the evaluation script will search for the best threshold
            all_predictions[example.qas_id] = best_non_null_entry.text

            all_nbest_json[example.qas_id] = nbest_json

        """
        if version_2_with_negative:
            with open(output_null_log_odds_file, "w") as writer:
                writer.write(json.dumps(scores_diff_json, indent=4) + "\n")

        with open(orig_data_file, "r", encoding='utf-8') as reader:
            orig_data = json.load(reader)["data"]
        """

        # All ids set as None so get rid of None Key
        all_predictions = all_predictions[None]
        all_nbest_json = all_nbest_json[None]

        return all_predictions, all_nbest_json

    def predict(
        self, query: str, context: str, n_best_size: int = 20, as_dict: bool = False
    ) -> Union[List[str], dict]:
        """Predicts top_n answer spans of query in regards to context

        Args:
            query: The question
            context: The context of which the question is asking
            top_n: The top n answers returned
            as_dict: Returns answer in dict format if True

        Returns:
            Either a list of string answers or a dict of the results
        """
        self._load() if not self.pretrained_model or not self.pretrained_tokenizer else None

        # Load and Evaluate Context Queries
        dataset, examples, features = self._load_one_query(query, context)
        eval_sampler = SequentialSampler(dataset)
        eval_dataloader = DataLoader(dataset, sampler=eval_sampler, batch_size=5)
        all_results = []
        for batch in eval_dataloader:
            self.pretrained_model.eval()
            batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                # BERT XLM XLNET DIFFERENCE
                inputs = {
                    "input_ids": batch[0],
                    "attention_mask": batch[1],
                    "token_type_ids": batch[2],
                    "cls_index": batch[4],
                    "p_mask": batch[5],
                }
                example_indices = batch[3]
                outputs = self.pretrained_model(**inputs)

            for i, example_index in enumerate(example_indices):
                eval_feature = features[example_index.item()]
                unique_id = int(eval_feature.unique_id)
                # BERT XLM XLNET DIFFERENCE
                result = RawResultExtended(
                    unique_id=unique_id,
                    start_top_log_probs=self._to_list(outputs[0][i]),
                    start_top_index=self._to_list(outputs[1][i]),
                    end_top_log_probs=self._to_list(outputs[2][i]),
                    end_top_index=self._to_list(outputs[3][i]),
                    cls_logits=self._to_list(outputs[4][i]),
                )
                all_results.append(result)

        # Obtain Concrete Predictions
        all_predictions, all_nbest_json = self._produce_concrete_predictions(
            examples, features, all_results, n_best_size=n_best_size
        )
        return all_predictions, all_nbest_json


class XLMQuestionAnsweringModel(QuestionAnsweringModel):
    def __init__(self):
        self.config = XLMConfig
        self.tokenizer = XLMTokenizer
        self.model = XLMForQuestionAnswering
        self.model_names = list(self.config.pretrained_config_archive_map.keys())

        # Post Load
        self.pretrained_config = None
        self.pretrained_tokenizer = None
        self.pretrained_model = None

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def _to_list(self, tensor: torch.Tensor) -> List[float]:
        return tensor.detach().cpu().tolist()

    def _load(self) -> None:
        print("Loading Pretrained XLNet Question Answering Model...")
        model_name = "xlm-mlm-en-2048"
        self.pretrained_config = self.config.from_pretrained("xlm-mlm-en-2048")
        tokenizer = self.tokenizer.from_pretrained(
            "xlm-mlm-en-2048", do_lower_case=False
        )
        self.pretrained_tokenizer = tokenizer

        model = self.model.from_pretrained(
            model_name,
            from_tf=bool(".ckpt" in model_name),
            config=self.pretrained_config,
        )
        self.pretrained_model = model
        self.pretrained_model.to(self.device)

    def _load_one_query(
        self, query: str, context: str, output_examples=True
    ) -> Union[TensorDataset, List[SquadExample], List[InputFeatures]]:
        # Create doc_tokens for SquadExample with one query and context

        def is_whitespace(c):
            if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
                return True
            return False

        # Create doc_tokens
        doc_tokens = []
        prev_is_whitespace = True
        for c in context:
            if is_whitespace(c):
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    doc_tokens.append(c)
                else:
                    doc_tokens[-1] += c
                prev_is_whitespace = False

        # Create SquadExample
        examples = []
        example = SquadExample(
            qas_id=None,
            question_text=query,
            doc_tokens=doc_tokens,
            orig_answer_text=None,
            start_position=None,
            end_position=None,
            is_impossible=False,
        )
        examples.append(example)

        # Convert to features
        features = convert_examples_to_features(
            examples=examples,
            tokenizer=self.pretrained_tokenizer,
            max_seq_length=384,
            doc_stride=128,
            max_query_length=64,
            is_training=False,
        )

        # Convert to Tensors and build dataset
        all_input_ids = torch.tensor([f.input_ids for f in features], dtype=torch.long)
        all_input_mask = torch.tensor(
            [f.input_mask for f in features], dtype=torch.long
        )
        all_segment_ids = torch.tensor(
            [f.segment_ids for f in features], dtype=torch.long
        )
        all_cls_index = torch.tensor([f.cls_index for f in features], dtype=torch.long)
        all_p_mask = torch.tensor([f.p_mask for f in features], dtype=torch.float)

        all_example_index = torch.arange(all_input_ids.size(0), dtype=torch.long)
        dataset = TensorDataset(
            all_input_ids,
            all_input_mask,
            all_segment_ids,
            all_example_index,
            all_cls_index,
            all_p_mask,
        )

        if output_examples:
            return dataset, examples, features
        return dataset

    def _produce_concrete_predictions(
        self,
        all_examples,
        all_features,
        all_results,
        n_best_size=10,
        max_answer_length=30,
        verbose_logging=False,
    ):

        start_n_top = self.pretrained_model.config.start_n_top
        end_n_top = self.pretrained_model.config.end_n_top
        tokenizer = self.pretrained_tokenizer

        _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "PrelimPrediction",
            [
                "feature_index",
                "start_index",
                "end_index",
                "start_log_prob",
                "end_log_prob",
            ],
        )

        _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "NbestPrediction", ["text", "start_log_prob", "end_log_prob"]
        )

        example_index_to_features = collections.defaultdict(list)
        for feature in all_features:
            example_index_to_features[feature.example_index].append(feature)

        unique_id_to_result = {}
        for result in all_results:
            unique_id_to_result[result.unique_id] = result

        all_predictions = collections.OrderedDict()
        all_nbest_json = collections.OrderedDict()
        scores_diff_json = collections.OrderedDict()

        for (example_index, example) in enumerate(all_examples):
            features = example_index_to_features[example_index]

            prelim_predictions = []
            # keep track of the minimum score of null start+end of position 0
            score_null = 1000000  # large and positive

            for (feature_index, feature) in enumerate(features):
                result = unique_id_to_result[feature.unique_id]

                cur_null_score = result.cls_logits

                # if we could have irrelevant answers, get the min score of irrelevant
                score_null = min(score_null, cur_null_score)

                for i in range(start_n_top):
                    for j in range(end_n_top):
                        start_log_prob = result.start_top_log_probs[i]
                        start_index = result.start_top_index[i]

                        j_index = i * end_n_top + j

                        end_log_prob = result.end_top_log_probs[j_index]
                        end_index = result.end_top_index[j_index]

                        # We could hypothetically create invalid predictions, e.g., predict
                        # that the start of the span is in the question. We throw out all
                        # invalid predictions.
                        if start_index >= feature.paragraph_len - 1:
                            continue
                        if end_index >= feature.paragraph_len - 1:
                            continue

                        if not feature.token_is_max_context.get(start_index, False):
                            continue
                        if end_index < start_index:
                            continue
                        length = end_index - start_index + 1
                        if length > max_answer_length:
                            continue

                        prelim_predictions.append(
                            _PrelimPrediction(
                                feature_index=feature_index,
                                start_index=start_index,
                                end_index=end_index,
                                start_log_prob=start_log_prob,
                                end_log_prob=end_log_prob,
                            )
                        )

            prelim_predictions = sorted(
                prelim_predictions,
                key=lambda x: (x.start_log_prob + x.end_log_prob),
                reverse=True,
            )

            seen_predictions = {}
            nbest = []
            for pred in prelim_predictions:
                if len(nbest) >= n_best_size:
                    break
                feature = features[pred.feature_index]

                # XLNet un-tokenizer
                # Let's keep it simple for now and see if we need all this later.
                #
                # tok_start_to_orig_index = feature.tok_start_to_orig_index
                # tok_end_to_orig_index = feature.tok_end_to_orig_index
                # start_orig_pos = tok_start_to_orig_index[pred.start_index]
                # end_orig_pos = tok_end_to_orig_index[pred.end_index]
                # paragraph_text = example.paragraph_text
                # final_text = paragraph_text[start_orig_pos: end_orig_pos + 1].strip()

                # Previously used Bert untokenizer
                tok_tokens = feature.tokens[pred.start_index : (pred.end_index + 1)]
                orig_doc_start = feature.token_to_orig_map[pred.start_index]
                orig_doc_end = feature.token_to_orig_map[pred.end_index]
                orig_tokens = example.doc_tokens[orig_doc_start : (orig_doc_end + 1)]
                tok_text = tokenizer.convert_tokens_to_string(tok_tokens)

                # Clean whitespace
                tok_text = tok_text.strip()
                tok_text = " ".join(tok_text.split())
                orig_text = " ".join(orig_tokens)

                final_text = get_final_text(
                    tok_text,
                    orig_text,
                    tokenizer,  # .do_lower_case, (a XLM problem?)
                    verbose_logging,
                )

                if final_text in seen_predictions:
                    continue

                seen_predictions[final_text] = True

                nbest.append(
                    _NbestPrediction(
                        text=final_text,
                        start_log_prob=pred.start_log_prob,
                        end_log_prob=pred.end_log_prob,
                    )
                )

            # In very rare edge cases we could have no valid predictions. So we
            # just create a nonce prediction in this case to avoid failure.
            if not nbest:
                nbest.append(
                    _NbestPrediction(text="", start_log_prob=-1e6, end_log_prob=-1e6)
                )

            total_scores = []
            best_non_null_entry = None
            for entry in nbest:
                total_scores.append(entry.start_log_prob + entry.end_log_prob)
                if not best_non_null_entry:
                    best_non_null_entry = entry

            probs = _compute_softmax(total_scores)

            nbest_json = []
            for (i, entry) in enumerate(nbest):
                output = collections.OrderedDict()
                output["text"] = entry.text
                output["probability"] = probs[i]
                output["start_log_prob"] = entry.start_log_prob
                output["end_log_prob"] = entry.end_log_prob
                nbest_json.append(output)

            assert len(nbest_json) >= 1
            assert best_non_null_entry is not None

            score_diff = score_null
            scores_diff_json[example.qas_id] = score_diff
            # note(zhiliny): always predict best_non_null_entry
            # and the evaluation script will search for the best threshold
            all_predictions[example.qas_id] = best_non_null_entry.text

            all_nbest_json[example.qas_id] = nbest_json

        """
        if version_2_with_negative:
            with open(output_null_log_odds_file, "w") as writer:
                writer.write(json.dumps(scores_diff_json, indent=4) + "\n")

        with open(orig_data_file, "r", encoding='utf-8') as reader:
            orig_data = json.load(reader)["data"]
        """

        # All ids set as None so get rid of None Key
        all_predictions = all_predictions[None]
        all_nbest_json = all_nbest_json[None]

        return all_predictions, all_nbest_json

    def predict(
        self, query: str, context: str, n_best_size: int = 20, as_dict: bool = False
    ) -> Union[List[str], dict]:
        """Predicts top_n answer spans of query in regards to context

        Args:
            query: The question
            context: The context of which the question is asking
            top_n: The top n answers returned
            as_dict: Returns answer in dict format if True

        Returns:
            Either a list of string answers or a dict of the results
        """
        self._load() if not self.pretrained_model or not self.pretrained_tokenizer else None

        # Load and Evaluate Context Queries
        dataset, examples, features = self._load_one_query(query, context)
        eval_sampler = SequentialSampler(dataset)
        eval_dataloader = DataLoader(dataset, sampler=eval_sampler, batch_size=5)
        all_results = []
        for batch in eval_dataloader:
            self.pretrained_model.eval()
            batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                # BERT XLM XLNET DIFFERENCE
                inputs = {
                    "input_ids": batch[0],
                    "attention_mask": batch[1],
                    "token_type_ids": batch[2],
                    "cls_index": batch[4],
                    "p_mask": batch[5],
                }
                example_indices = batch[3]
                outputs = self.pretrained_model(**inputs)

            for i, example_index in enumerate(example_indices):
                eval_feature = features[example_index.item()]
                unique_id = int(eval_feature.unique_id)
                # BERT XLM XLNET DIFFERENCE
                result = RawResultExtended(
                    unique_id=unique_id,
                    start_top_log_probs=self._to_list(outputs[0][i]),
                    start_top_index=self._to_list(outputs[1][i]),
                    end_top_log_probs=self._to_list(outputs[2][i]),
                    end_top_index=self._to_list(outputs[3][i]),
                    cls_logits=self._to_list(outputs[4][i]),
                )
                all_results.append(result)

        # Obtain Concrete Predictions
        all_predictions, all_nbest_json = self._produce_concrete_predictions(
            examples, features, all_results, n_best_size=n_best_size
        )
        return all_predictions, all_nbest_json



# coding=utf-8
# This file uses code from the language modeling examples in the huggingface Transformer's repo

import os
import logging
import math
from typing import Dict, Union
from pathlib import Path

import torch
from torch.utils.data import Dataset
from transformers import (
    AutoTokenizer,
    AutoModelWithLMHead,
    TrainingArguments,
    Trainer,
    TextDataset,
    DataCollatorForLanguageModeling,
    # TODO: For XLNet, will be available in Transformers release 3.0.2+
    # DataCollatorForPermutationLanguageModeling,
    LineByLineTextDataset,
)

logger = logging.getLogger(__name__)


class LMFineTuner:
    """
     A Language Model Fine Tuner object you can set language model configurations and then train and evaluate

    Usage:

    ```python
    >>> finetuner = adaptnlp.LMFineTuner()
    >>> finetuner.train()
    ```

    **Parameters:**

    * **model_name_or_path** - The model checkpoint for weights initialization. Leave None if you want to train a model from scratch.
    """

    def __init__(
        self,
        model_name_or_path="bert-base-cased",
    ):

        logger.info(
            "This is the new updated `LMFineTuner` class object for 0.2.0+. If you're looking for `LMFineTuner` from <=0.1.6, you can instantiate it with LMFineTunerManual"
        )
        # Load model and tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path, use_fast=True
        )
        # TODO: AutoModelWithLMHead deprecated, update to causal, mask, or seq2seq
        self.model = AutoModelWithLMHead.from_pretrained(model_name_or_path)
        self.trainer = None

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

    def train(
        self,
        training_args: TrainingArguments,
        train_file: Union[str, Path],
        eval_file: Union[str, Path],
        line_by_line: bool = False,
        mlm: bool = False,
        mlm_probability: float = 0.15,
        plm_probability: float = 1 / 6,
        max_span_length: int = 5,
        block_size: int = -1,
        overwrite_cache: bool = False,
    ):
        """Train and fine-tune the loaded language model

        * **train_file** - The input training data file (a text file).
        * **eval_file** - An optional input evaluation data file to evaluate the perplexity on (a text file).
        * **line_by_line** - Whether distinct lines of text in the dataset are to be handled as distinct sequences.
        * **mlm** - Train with masked-language modeling loss instead of language modeling.
        * **mlm_probability** - Ratio of tokens to mask for masked language modeling loss
        * **plm_probability** - Ratio of length of a span of masked tokens to surrounding context length for permutation language modeling.
        * **max_span_length** - Maximum length of a span of masked tokens for permutation language modeling.
        * **block_size** - Optional input sequence length after tokenization.
                            The training dataset will be truncated in block of this size for training."
                            `-1` will default to the model max input length for single sentence inputs (take into account special tokens).
        * **overwrite_cache** - Overwrite the cached training and evaluation sets
        """

        # Setup logging
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=logging.INFO if training_args.local_rank in [-1, 0] else logging.WARN,
        )
        logger.warning(
            f"""Process rank: {training_args.local_rank},
                device: {training_args.device},
                n_gpu: {training_args.n_gpu},
                distributed training: {bool(training_args.local_rank != -1)},
                16-bits training: {training_args.fp16}
            """
        )
        logger.info(f"Training/evaluation parameters: {training_args.to_json_string()}")

        # Check if masked language model or not
        if (
            self.model.config.model_type
            in ["bert", "roberta", "distilbert", "camembert"]
            and not mlm
        ):
            raise ValueError(
                """BERT and RoBERTa-like models do not have LM heads but masked LM heads. They must be run with
                mlm set as True(masked language modeling)."""
            )

        # Check block size for Dataset
        if block_size <= 0:
            block_size = self.tokenizer.max_len
        else:
            block_size = min(block_size, self.tokenizer.max_len)

        # Get datasets
        train_dataset = self._get_dataset(
            file_path=train_file,
            line_by_line=line_by_line,
            block_size=block_size,
            overwrite_cache=overwrite_cache,
        )
        eval_dataset = self._get_dataset(
            file_path=eval_file,
            line_by_line=line_by_line,
            block_size=block_size,
            overwrite_cache=overwrite_cache,
        )
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset

        # Get Collator
        # TODO: DataCollatorForPermutationLanguageModeling not availbe until release 3.0.2+
        if self.model.config.model_type == "xlnet":
            logger.info("Cannot currently finetune XLNet model")
            raise ValueError(
                "Use another language model besides XLNet for LM finetuning"
            )
            """
            data_collator = DataCollatorForPermutationLanguageModeling(
            tokenizer=self.tokenizer,
            plm_probability=plm_probability,
            max_span_length=max_span_length,
            )
            """
        else:
            data_collator = DataCollatorForLanguageModeling(
                tokenizer=self.tokenizer, mlm=mlm, mlm_probability=mlm_probability
            )

        # Initialize Trainer
        self.trainer = Trainer(
            model=self.model,
            args=training_args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
        )

        # Train and serialize
        self.trainer.train()
        self.trainer.save_model()
        self.tokenizer.save_pretrained(training_args.output_dir)

    def evaluate(self) -> Dict[str, float]:

        if not self.trainer:
            logger.info(
                "No trainer loaded, you should probably run `LMFineTuner.train(...)` first"
            )
            return None
        results = {}

        logger.info("*** Evaluate ***")

        eval_output = self.trainer.evaluate()

        perplexity = math.exp(eval_output["eval_loss"])
        result = {"perplexity": perplexity}

        output_eval_file = os.path.join(
            self.trainer.args.output_dir, "eval_results_lm.txt"
        )

        with open(output_eval_file, "w") as writer:
            logger.info("***** Eval results *****")
            for key in sorted(result.keys()):
                logger.info("  %s = %s", key, str(result[key]))
                writer.write("%s = %s\n" % (key, str(result[key])))
        results.update(result)

        return results

    def _get_dataset(
        self,
        file_path: str,
        line_by_line: bool,
        block_size: int,
        overwrite_cache: bool,
    ) -> Dataset:
        if line_by_line:
            return LineByLineTextDataset(
                tokenizer=self.tokenizer, file_path=file_path, block_size=block_size
            )
        else:
            return TextDataset(
                tokenizer=self.tokenizer,
                file_path=file_path,
                block_size=block_size,
                overwrite_cache=overwrite_cache,
            )



import pkg_resources
from pathlib import Path

# Easy Modules
from .embeddings import (
    EasyWordEmbeddings,
    EasyStackedEmbeddings,
    EasyDocumentEmbeddings,
)
from .token_classification import EasyTokenTagger
from .sequence_classification import (
    EasySequenceClassifier,
    TransformersSequenceClassifier,
    FlairSequenceClassifier,
)
from .question_answering import EasyQuestionAnswering, TransformersQuestionAnswering
from .summarization import EasySummarizer, TransformersSummarizer
from .translation import EasyTranslator, TransformersTranslator
from .text_generation import EasyTextGenerator, TransformersTextGenerator
from .language_model import LMFineTuner

# Training and Fine-tuning Modules
# TODO: Deprecating in 0.3.0+
from .training import SequenceClassifierTrainer
from .transformers.finetuning import LMFineTunerManual

# global variable like flair's: cache_root
cache_root = Path(Path.home(), ".adaptnlp")

__version__ = (
    pkg_resources.resource_string("adaptnlp", "VERSION.txt").decode("UTF-8").strip()
)

__all__ = [
    "__version__",
    "EasyWordEmbeddings",
    "EasyStackedEmbeddings",
    "EasyDocumentEmbeddings",
    "EasyTokenTagger",
    "EasySequenceClassifier",
    "FlairSequenceClassifier",
    "TransformersSequenceClassifier",
    "EasyQuestionAnswering",
    "TransformersQuestionAnswering",
    "EasySummarizer",
    "TransformersSummarizer",
    "EasyTranslator",
    "TransformersTranslator",
    "EasyTextGenerator",
    "TransformersTextGenerator",
    "SequenceClassifierTrainer",
    "LMFineTuner",
    "LMFineTunerManual",
]



import logging
from typing import List, Dict, Union
from collections import defaultdict

import torch
from torch.utils.data import TensorDataset, DataLoader

from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    PreTrainedTokenizer,
    PreTrainedModel,
    T5ForConditionalGeneration,
)

from tqdm import tqdm

from adaptnlp.model import AdaptiveModel

logger = logging.getLogger(__name__)


class TransformersTranslator(AdaptiveModel):
    """Adaptive model for Transformer's Conditional Generation or Language Models (Transformer's T5 and Bart
    conditional generation models have a language modeling head)

    Usage:
    ```python
    >>> translator = TransformersTranslator.load("transformers-translator-model")
    >>> translator.predict(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers
    * **model** - A transformers Conditional Generation (Bart or T5) or Language model
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        self.model = model

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this classifier

        * **model_name_or_path** - A key string of one of Transformer's pre-trained translator Model
        """
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path)
        translator = cls(tokenizer, model)
        return translator

    def predict(
        self,
        text: Union[List[str], str],
        t5_prefix: str = "translate English to German",
        mini_batch_size: int = 32,
        num_beams: int = 1,
        min_length: int = 0,
        max_length: int = 128,
        early_stopping: bool = True,
        **kwargs,
    ) -> List[str]:
        """Predict method for running inference using the pre-trained sequence classifier model.  Keyword arguments
        for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well.

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **t5_prefix**(Optional) - The pre-appended prefix for the specificied task. Only in use for T5-type models.
        * **mini_batch_size** - Mini batch size
        * **num_beams** - Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search.  Default to 1.
        * **min_length** -  The min length of the sequence to be generated. Default to 0
        * **max_length** - The max length of the sequence to be generated. Between min_length and infinity. Default to 128
        * **early_stopping** - if set to True beam search is stopped when at least num_beams sentences finished per batch.
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers `PreTrainedModel.generate()` method
        """
        with torch.no_grad():

            # Make all inputs lists
            if isinstance(text, str):
                text = [text]

            # T5 requires "translate: " precursor text for pre-trained translator
            if isinstance(self.model, T5ForConditionalGeneration):
                text = [f"{t5_prefix}: {t}" for t in text]

            dataset = self._tokenize(text)
            dataloader = DataLoader(dataset, batch_size=mini_batch_size)
            translations = []

            logger.info(f"Running translator on {len(dataset)} text sequences")
            logger.info(f"Batch size = {mini_batch_size}")
            for batch in tqdm(dataloader, desc="Translating"):
                self.model.eval()
                batch = tuple(t.to(self.device) for t in batch)

                if len(batch) == 3:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_masks": batch[1],
                        "token_type_ids": batch[2],
                    }
                else:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_masks": batch[1],
                    }
                outputs = self.model.generate(
                    inputs["input_ids"],
                    num_beams=num_beams,
                    min_length=min_length,
                    max_length=max_length,
                    early_stopping=early_stopping,
                    **kwargs,
                )

                for o in outputs:
                    translations.append(
                        [
                            self.tokenizer.decode(
                                o,
                                skip_special_tokens=True,
                                clean_up_tokenization_spaces=False,
                            )
                        ].pop()
                    )

        return translations

    def _tokenize(self, text: Union[List[str], str]) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with text """

        tokenized_text = self.tokenizer.batch_encode_plus(
            text,
            return_tensors="pt",
            max_length=512,
            padding="max_length",
            add_special_tokens=True,
        )

        # Bart doesn't use `token_type_ids`
        dataset = TensorDataset(
            tokenized_text["input_ids"],
            tokenized_text["attention_mask"],
        )
        return dataset

    def train(
        self,
    ):
        raise NotImplementedError

    def evaluate(
        self,
    ):
        raise NotImplementedError


class EasyTranslator:
    """Translation Module

    Usage:

    ```python
    >>> translator = EasyTranslator()
    >>> translator.translate(text="translate this text", model_name_or_path="t5-small")
    ```

    """

    def __init__(self):
        self.translators: Dict[AdaptiveModel] = defaultdict(bool)

    def translate(
        self,
        text: Union[List[str], str],
        model_name_or_path: str = "t5-small",
        t5_prefix: str = "translate English to German",
        mini_batch_size: int = 32,
        num_beams: int = 1,
        min_length: int = 0,
        max_length: int = 128,
        early_stopping: bool = True,
        **kwargs,
    ) -> List[str]:
        """Predict method for running inference using the pre-trained sequence classifier model. Keyword arguments
        for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well.

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **model_name_or_path** - A String model id or path to a pre-trained model repository or custom trained model directory
        * **t5_prefix**(Optional) - The pre-appended prefix for the specificied task. Only in use for T5-type models.
        * **mini_batch_size** - Mini batch size
        * **num_beams** - Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search.  Default to 1.
        * **min_length** -  The min length of the sequence to be generated. Default to 0
        * **max_length** - The max length of the sequence to be generated. Between min_length and infinity. Default to 128
        * **early_stopping** - if set to True beam search is stopped when at least num_beams sentences finished per batch.
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers `PreTrainedModel.generate()` method
        """
        if not self.translators[model_name_or_path]:
            self.translators[model_name_or_path] = TransformersTranslator.load(
                model_name_or_path
            )

        translator = self.translators[model_name_or_path]
        return translator.predict(
            text=text,
            t5_prefix=t5_prefix,
            mini_batch_size=mini_batch_size,
            num_beams=num_beams,
            min_length=min_length,
            max_length=max_length,
            early_stopping=early_stopping,
            **kwargs,
        )



import logging
from typing import List, Dict, Union, Tuple, Callable
from collections import defaultdict
from pathlib import Path

import torch
from torch import nn
from torch.utils.data import TensorDataset, DataLoader
import datasets
from datasets import ClassLabel
from sklearn.metrics import accuracy_score, precision_recall_fscore_support
from flair.data import Sentence, DataPoint
from flair.models import TextClassifier
from transformers import (
    AutoTokenizer,
    AutoModelForSequenceClassification,
    PreTrainedTokenizer,
    PreTrainedModel,
    BertPreTrainedModel,
    DistilBertPreTrainedModel,
    XLMPreTrainedModel,
    XLNetPreTrainedModel,
    ElectraPreTrainedModel,
    BertForSequenceClassification,
    XLNetForSequenceClassification,
    AlbertForSequenceClassification,
    TrainingArguments,
    Trainer,
)

from tqdm import tqdm


from adaptnlp.model import AdaptiveModel

logger = logging.getLogger(__name__)


class TransformersSequenceClassifier(AdaptiveModel):
    """Adaptive model for Transformer's Sequence Classification Model

    Usage:
    ```python
    >>> classifier = TransformersSequenceClassifier.load("transformers-sc-model")
    >>> classifier.predict(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers
    * **model** - A transformers Sequence Classsifciation model
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        self.model = model

        # Load empty trainer
        self.trainer = None

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this classifier

        * **model_name_or_path** - A key string of one of Transformer's pre-trained Sequence Classifier Model
        """
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=True)
        model = AutoModelForSequenceClassification.from_pretrained(model_name_or_path)
        classifier = cls(tokenizer, model)
        return classifier

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Predict method for running inference using the pre-trained sequence classifier model

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **mini_batch_size** - Mini batch size
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers classifier
        """
        id2label = self.model.config.id2label
        sentences = text
        results: List[Sentence] = []

        with torch.no_grad():
            if not sentences:
                return sentences

            if isinstance(sentences, DataPoint) or isinstance(sentences, str):
                sentences = [sentences]

            # filter empty sentences
            if isinstance(sentences[0], Sentence):
                sentences = [sentence for sentence in sentences if len(sentence) > 0]
            if len(sentences) == 0:
                return sentences

            # reverse sort all sequences by their length
            rev_order_len_index = sorted(
                range(len(sentences)), key=lambda k: len(sentences[k]), reverse=True
            )
            original_order_index = sorted(
                range(len(rev_order_len_index)), key=lambda k: rev_order_len_index[k]
            )

            reordered_sentences: List[Union[DataPoint, str]] = [
                sentences[index] for index in rev_order_len_index
            ]
            # Turn all Sentence objects into strings
            if isinstance(reordered_sentences[0], Sentence):
                str_reordered_sentences = [
                    sentence.to_original_text() for sentence in sentences
                ]
            else:
                str_reordered_sentences = reordered_sentences

            # Tokenize and get dataset
            dataset = self._tokenize(str_reordered_sentences)
            dataloader = DataLoader(dataset, batch_size=mini_batch_size)
            predictions: List[Tuple[str, float]] = []

            logger.info(f"Running prediction on {len(dataset)} text sequences")
            logger.info(f"Batch size = {mini_batch_size}")
            for batch in tqdm(dataloader, desc="Predicting text"):
                self.model.eval()
                batch = tuple(t.to(self.device) for t in batch)

                if len(batch) == 3:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_mask": batch[1],
                        "token_type_ids": batch[2],
                    }
                else:
                    inputs = {"input_ids": batch[0], "attention_mask": batch[1]}
                outputs = self.model(**inputs)
                logits = outputs[0]
                preds = torch.softmax(logits, dim=1).tolist()

                predictions += preds

            for text, pred in zip(str_reordered_sentences, predictions):
                # Initialize and assign labels to each class in each datapoint prediction
                text_sent = Sentence(text)
                for k, v in id2label.items():
                    text_sent.add_label(label_type="sc", value=v, score=pred[k])
                results.append(text_sent)

        # Order results back into original order
        results = [results[index] for index in original_order_index]

        return results

    def _tokenize(
        self, sentences: Union[List[Sentence], Sentence, List[str], str]
    ) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with them """

        # TODO: __call__ from tokenizer base class in the transformers library could automate/handle this
        tokenized_text = self.tokenizer.batch_encode_plus(
            sentences,
            return_tensors="pt",
            pad_to_max_length=True,
            add_special_tokens=True,
        )

        # Bart, XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use token_type_ids
        if isinstance(
            self.model,
            (
                BertForSequenceClassification,
                XLNetForSequenceClassification,
                AlbertForSequenceClassification,
            ),
        ):
            dataset = TensorDataset(
                tokenized_text["input_ids"],
                tokenized_text["attention_mask"],
                tokenized_text["token_type_ids"],
            )
        else:
            dataset = TensorDataset(
                tokenized_text["input_ids"], tokenized_text["attention_mask"]
            )

        return dataset

    def train(
        self,
        training_args: TrainingArguments,
        train_dataset: datasets.Dataset,
        eval_dataset: datasets.Dataset,
        text_col_nm: str = "text",
        label_col_nm: str = "label",
        compute_metrics: Callable = None,
    ) -> None:
        """Trains and/or finetunes the sequence classification model

        * **training_args** - Transformers `TrainingArguments` object model
        * **train_dataset** - Training `Dataset` class object from the datasets library
        * **eval_dataset** - Eval `Dataset` class object from the datasets library
        * **text_col_nm** - Name of the text feature column used as training data (Default "text")
        * **label_col_nm** - Name of the label feature column (Default "label")
        * **compute_metrics** - Custom metrics function callable for `transformers.Trainer`'s compute metrics
        * **return** - None
        """
        # Set default metrics if None
        if not compute_metrics:
            compute_metrics = self._default_metrics

        # Set datasets.Dataset label values in sequence classifier configuration
        ## Important NOTE: Updating configurations do not update the sequence classification head module layer
        ## We are manually initializing a new linear layer for the "new" labels being trained
        class_label = train_dataset.features[label_col_nm]
        config_data = {
            "num_labels": class_label.num_classes,
            "id2label": {v: n for v, n in enumerate(class_label.names)},
            "label2id": {n: v for v, n in enumerate(class_label.names)},
        }
        self.model.config.update(config_data)
        self._mutate_model_head(class_label=class_label)

        # Batch map datasets as torch tensors with tokenizer
        def tokenize(batch):
            return self.tokenizer(batch[text_col_nm], padding=True, truncation=True)

        train_dataset = train_dataset.map(
            tokenize, batch_size=len(train_dataset), batched=True
        )
        eval_dataset = eval_dataset.map(
            tokenize, batch_size=len(eval_dataset), batched=True
        )

        # Rename label col name to match model forward signature of "labels" or ["label","label_ids"] since these are addressed by the default collator from transformers
        train_dataset.rename_column_(
            original_column_name=label_col_nm, new_column_name="labels"
        )
        eval_dataset.rename_column_(
            original_column_name=label_col_nm, new_column_name="labels"
        )

        # Set format as torch tensors for training
        train_dataset.set_format(
            "torch", columns=["input_ids", "attention_mask", "labels"]
        )
        eval_dataset.set_format(
            "torch", columns=["input_ids", "attention_mask", "labels"]
        )

        # Instantiate transformers trainer
        self.trainer = Trainer(
            model=self.model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=compute_metrics,
        )

        # Train and serialize
        self.trainer.train()
        self.trainer.save_model()
        self.tokenizer.save_pretrained(training_args.output_dir)

    def evaluate(self) -> Dict[str, float]:
        """Evaluates model specified

        * **model_name_or_path** - The model name key or model path
        """
        if not self.trainer:
            logger.info("No trainer loaded, must run `classifier.train(...)` first")
            ValueError("Trainer not found, must run train() method")
        return self.trainer.evaluate()

    def _mutate_model_head(self, class_label: ClassLabel) -> None:
        """Manually intialize new linear layers for prediction heads on specific language models that we're trying to train on"""
        if isinstance(self.model, (BertPreTrainedModel, DistilBertPreTrainedModel)):
            self.model.classifier = nn.Linear(
                self.model.config.hidden_size, class_label.num_classes
            )
            self.model.num_labels = class_label.num_classes
        elif isinstance(self.model, XLMPreTrainedModel):
            self.model.num_labels = class_label.num_classes
        elif isinstance(self.model, XLNetPreTrainedModel):
            self.model.logits_proj = nn.Linear(
                self.model.config.d_model, class_label.num_classes
            )
            self.model.num_labels = class_label.num_classes
        elif isinstance(self.model, ElectraPreTrainedModel):
            self.model.num_labels = class_label.num_classes
        else:
            logger.info(f"Sorry, can not train on a model of type {type(self.model)}")

    # Setup default metrics for sequence classification training
    def _default_metrics(self, pred) -> Dict:
        labels = pred.label_ids
        preds = pred.predictions.argmax(-1)
        precision, recall, f1, _ = precision_recall_fscore_support(
            labels, preds, average=None
        )
        acc = accuracy_score(labels, preds)
        return {"accuracy": acc, "f1": f1, "precision": precision, "recall": recall}


class FlairSequenceClassifier(AdaptiveModel):
    """Adaptive Model for Flair's Sequence Classifier...very basic

    Usage:
    ```python
    >>> classifier = FlairSequenceClassifier.load("en-sentiment")
    >>> classifier.predict(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **model_name_or_path** - A key string of one of Flair's pre-trained Sequence Classifier Model
    """

    def __init__(self, model_name_or_path: str):
        self.classifier = TextClassifier.load(model_name_or_path)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading a constructing this classifier

        * **model_name_or_path** - A key string of one of Flair's pre-trained Sequence Classifier Model
        """
        classifier = cls(model_name_or_path)
        return classifier

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Predict method for running inference using the pre-trained sequence classifier model

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **mini_batch_size** - Mini batch size
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Flair classifier
        """
        if isinstance(text, (Sentence, str)):
            text = [text]
        if isinstance(text[0], str):
            text = [Sentence(s) for s in text]
        return self.classifier.predict(
            sentences=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )

    def train(self):
        pass

    def evaluate(self):
        pass


class EasySequenceClassifier:
    """Sequence classification models

    Usage:

    ```python
    >>> classifier = EasySequenceClassifier()
    >>> classifier.tag_text(text="text you want to label", model_name_or_path="en-sentiment")
    ```

    """

    def __init__(self):
        self.sequence_classifiers: Dict[AdaptiveModel] = defaultdict(bool)

    def tag_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        model_name_or_path: str = "en-sentiment",
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Tags a text sequence with labels the sequence classification models have been trained on

        * **text** - String, list of strings, `Sentence`, or list of `Sentence`s to be classified
        * **model_name_or_path** - The model name key or model path
        * **mini_batch_size** - The mini batch size for running inference
        * **&ast;&ast;kwargs** - (Optional) Keyword Arguments for Flair's `TextClassifier.predict()` method params
        **return** A list of Flair's `Sentence`'s
        """
        # Load Text Classifier Model and Pytorch Module into tagger dict
        if not self.sequence_classifiers[model_name_or_path]:
            """
            self.sequence_classifiers[model_name_or_path] = TextClassifier.load(
                model_name_or_path
            )
            """
            # TODO: Find an alternative model-check method like an `is_available(model_name_or_path)
            # Check whether this is a Transformers or Flair Sequence Classifier model we're loading
            try:
                self.sequence_classifiers[
                    model_name_or_path
                ] = FlairSequenceClassifier.load(model_name_or_path)
            except (NotADirectoryError, FileNotFoundError, IsADirectoryError):
                logger.info(
                    f"{model_name_or_path} not a valid Flair pre-trained model...checking transformers repo"
                )
                try:
                    self.sequence_classifiers[
                        model_name_or_path
                    ] = TransformersSequenceClassifier.load(model_name_or_path)
                except ValueError:
                    logger.info("Try transformers")
                    return [Sentence("")]

        classifier = self.sequence_classifiers[model_name_or_path]
        return classifier.predict(
            text=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )

    def tag_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Tags text with all labels from all sequence classification models

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        * **mini_batch_size** - The mini batch size for running inference
        * **&ast;&ast;kwargs** - (Optional) Keyword Arguments for Flair's `TextClassifier.predict()` method params
        * **return** - A list of Flair's `Sentence`'s
        """
        sentences = text
        for tagger_name in self.sequence_classifiers.keys():
            sentences = self.tag_text(
                sentences,
                model_name_or_path=tagger_name,
                mini_batch_size=mini_batch_size,
                **kwargs,
            )
        return sentences

    def train(
        self,
        training_args: TrainingArguments,
        train_dataset: Union[str, Path, datasets.Dataset],
        eval_dataset: Union[str, Path, datasets.Dataset],
        model_name_or_path: str = "bert-base-uncased",
        text_col_nm: str = "text",
        label_col_nm: str = "label",
        label_names: List[str] = None,
    ) -> None:
        """Trains and/or finetunes the sequence classification model

        * **model_name_or_path** - The model name key or model path
        * **training_args** - Transformers `TrainingArguments` object model
        * **train_dataset** - Training `Dataset` class object from the datasets library or path to CSV file (labels must be int values)
        * **eval_dataset** - Eval `Dataset` class object from the datasets library or path to CSV file (labels must be int values)
        * **text_col_nm** - Name of the text feature column used as training data (Default "text")
        * **label_col_nm** - Name of the label feature column (Default "label")
        * **label_names** - (Only when loading CSV) An ordered list of label strings with int label mapped to string via. index value
        * **return** - None
        """

        # Dynamically load sequence classifier
        if not self.sequence_classifiers[model_name_or_path]:
            try:
                self.sequence_classifiers[
                    model_name_or_path
                ] = TransformersSequenceClassifier.load(model_name_or_path)
            except ValueError:
                logger.info("Try transformers model")

        classifier = self.sequence_classifiers[model_name_or_path]

        # Check if csv filepath or `datasets.Dataset`
        if not isinstance(train_dataset, datasets.Dataset):
            train_dataset = self._csv2dataset(
                data_path=train_dataset,
                label_col_nm=label_col_nm,
                label_names=label_names,
            )
        if not isinstance(eval_dataset, datasets.Dataset):
            eval_dataset = self._csv2dataset(
                data_path=eval_dataset,
                label_col_nm=label_col_nm,
                label_names=label_names,
            )

        classifier.train(
            training_args=training_args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            text_col_nm=text_col_nm,
            label_col_nm=label_col_nm,
        )

    def release_model(
        self,
        model_name_or_path: str,
    ) -> None:
        """Unload model from classifier and empty cuda mem cache (may leave residual cache per pytorch documentation on torch.cuda.empty_cache())

        * **model_name_or_path** - The model name or key path that you want to unload and release memory from
        """
        if self.sequence_classifiers[model_name_or_path].trainer:
            del self.sequence_classifiers[model_name_or_path].trainer
        del self.sequence_classifiers[model_name_or_path]
        torch.cuda.empty_cache()

    def evaluate(
        self,
        model_name_or_path: str = "bert-base-uncased",
    ) -> Dict[str, float]:
        """Evaluates model specified

        * **model_name_or_path** - The model name key or model path
        """

        # Dynamically load sequence classifier
        if not self.sequence_classifiers[model_name_or_path]:
            try:
                self.sequence_classifiers[
                    model_name_or_path
                ] = TransformersSequenceClassifier.load(model_name_or_path)
            except ValueError:
                logger.info("Try transformers model")

        classifier = self.sequence_classifiers[model_name_or_path]

        return classifier.evaluate()

    def _csv2dataset(
        self,
        data_path: Union[str, Path],
        label_col_nm: str,
        label_names: List[str],
    ) -> datasets.Dataset:
        """Loads CSV path as an datasets.Dataset for downstream use"""
        if not label_names:
            raise ValueError(
                "Must pass in `label_names` parameter for training when loading in CSV datasets."
            )
        class_label = datasets.ClassLabel(
            num_classes=len(label_names), names=label_names
        )
        dataset = datasets.load_dataset("csv", data_files=data_path)
        dataset = dataset["train"]
        dataset.features[label_col_nm] = class_label
        return dataset



import logging
from typing import List, Dict, Union
from collections import defaultdict

import torch
from torch.utils.data import TensorDataset, DataLoader

from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    PreTrainedTokenizer,
    PreTrainedModel,
)

from tqdm import tqdm

from adaptnlp.model import AdaptiveModel

logger = logging.getLogger(__name__)


class TransformersTextGenerator(AdaptiveModel):
    """Adaptive model for Transformer's Language Models

    Usage:
    ```python
    >>> generator = TransformersTextGenerator.load("gpt2")
    >>> generator.generate(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers
    * **model** - A transformers Language model
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        self.model = model

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this Model

        * **model_name_or_path** - A key string of one of Transformer's pre-trained Language Model
        """
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, pad_token="<PAD>")
        model = AutoModelForCausalLM.from_pretrained(model_name_or_path)
        generator = cls(tokenizer, model)
        return generator

    def predict(
        self,
        text: Union[List[str], str],
        mini_batch_size: int = 32,
        num_tokens_to_produce: int = 50,
        **kwargs,
    ) -> List[str]:
        """Predict method for running inference using the pre-trained sequence classifier model.  Keyword arguments
        for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well.

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **mini_batch_size** - Mini batch size
        * **num_tokens_to_produce** - Number of tokens you want to generate
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers `PreTrainedModel.generate()` method
        """
        with torch.no_grad():

            # Make all inputs lists
            if isinstance(text, str):
                text = [text]

            dataset = self._tokenize(text)
            dataloader = DataLoader(dataset, batch_size=mini_batch_size)
            results = []

            logger.info(f"Running text generator on {len(dataset)} text sequences")
            logger.info(f"Batch size = {mini_batch_size}")
            for batch in tqdm(dataloader, desc="Generating"):
                self.model.eval()
                batch = tuple(t.to(self.device) for t in batch)

                if len(batch) == 3:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_masks": batch[1],
                        "token_type_ids": batch[2],
                    }
                else:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_masks": batch[1],
                    }
                # model.generate() does not have batch inference implemented yet
                generated_text = self._batch_generate(
                    inputs=inputs,
                    seq_len=batch[0].shape[1],
                    num_tokens_to_produce=num_tokens_to_produce,
                )
                results += generated_text

        return results

    def _tokenize(self, text: Union[List[str], str]) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with text """

        tokenized_text = self.tokenizer.batch_encode_plus(
            text,
            return_tensors="pt",
            padding="longest",
        )

        dataset = TensorDataset(
            tokenized_text["input_ids"],
            tokenized_text["attention_mask"],
        )

        return dataset

    def _batch_generate(
        self, inputs: Dict, seq_len: int, num_tokens_to_produce: int
    ) -> List[str]:
        """Generates text data with varying text sizes"""
        input_ids = inputs["input_ids"]
        attn_mask = inputs["attention_masks"]

        pad_token_id = self.tokenizer.pad_token_id
        eos_token_id = self.tokenizer.eos_token_id
        eos_not_in_sents = torch.ones(input_ids.shape[0]).long().to(self.device)

        # we need to get the token ids of the last non-padded value
        last_non_masked_idx = torch.sum(attn_mask, dim=1) - 1
        start_idx = (
            (last_non_masked_idx)
            .view(-1, 1)
            .repeat(1, self.tokenizer.vocab_size)
            .unsqueeze(1)
        )

        # get correct position ids
        position_ids = torch.tensor(
            [list(range(seq_len)) for i in range(input_ids.shape[0])]
        ).to(self.device)
        for i, position_ids_slice in enumerate(position_ids):
            position_ids_slice[last_non_masked_idx[i] :] = position_ids_slice[
                last_non_masked_idx[i]
            ]

        for step in range(num_tokens_to_produce):
            outputs = self.model(
                input_ids, attention_mask=attn_mask, position_ids=position_ids
            )

            # in the first decoding step, we want to use the 'real' last position for each sentence
            if step == 0:
                next_token_logits = outputs[0].gather(1, start_idx).squeeze(1)
            else:
                next_token_logits = outputs[0][:, -1, :]

            next_tokens = torch.argmax(next_token_logits, dim=-1)

            # this updates which sentences have not seen an <EOS> token so far
            # if one <EOS> token was seen the sentence is finished
            eos_not_in_sents.mul_(next_tokens.ne(eos_token_id).long())

            # either append a padding token here if <EOS> has been seen or append next token
            tokens_to_add = next_tokens * (eos_not_in_sents) + pad_token_id * (
                1 - eos_not_in_sents
            )

            # Update input_ids, attn_mask and position_ids
            input_ids = torch.cat([input_ids, tokens_to_add.unsqueeze(-1)], dim=-1)
            attn_mask = torch.cat(
                [attn_mask, torch.ones((attn_mask.shape[0], 1)).long().to(self.device)],
                dim=1,
            )
            position_ids = torch.cat(
                [position_ids, (position_ids[:, -1] + 1).unsqueeze(-1)], dim=1
            )

        return [
            self.tokenizer.decode(output, skip_special_tokens=True)
            for output in input_ids
        ]

    def train(
        self,
    ):
        raise NotImplementedError

    def evaluate(
        self,
    ):
        raise NotImplementedError


class EasyTextGenerator:
    """Text Generation Module

    Usage:

    ```python
    >>> generator = EasyGenerator()
    >>> generator.generate(text="generate from this text", num_tokens_to_produce=50)
    ```

    """

    def __init__(self):
        self.generators: Dict[AdaptiveModel] = defaultdict(bool)

    def generate(
        self,
        text: Union[List[str], str],
        model_name_or_path: str = "gpt2",
        mini_batch_size: int = 32,
        num_tokens_to_produce: int = 50,
        **kwargs,
    ) -> List[str]:
        """Predict method for running inference using the pre-trained sequence classifier model. Keyword arguments
        for parameters of the method `Transformers.PreTrainedModel.generate()` can be used as well.

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **model_name_or_path** - A String model id or path to a pre-trained model repository or custom trained model directory
        * **mini_batch_size** - Mini batch size
        * **num_tokens_to_produce** - Number of tokens you want to generate
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers `PreTrainedModel.generate()` method
        """
        if not self.generators[model_name_or_path]:
            self.generators[model_name_or_path] = TransformersTextGenerator.load(
                model_name_or_path
            )

        generator = self.generators[model_name_or_path]
        return generator.predict(
            text=text,
            mini_batch_size=mini_batch_size,
            num_tokens_to_produce=num_tokens_to_produce,
            **kwargs,
        )



import logging
from typing import List, Dict, Union
from collections import defaultdict

import numpy as np

import torch
from torch.utils.data import TensorDataset, DataLoader

from flair.data import Sentence
from flair.models import SequenceTagger
from transformers import (
    AutoTokenizer,
    AutoModelForTokenClassification,
    PreTrainedTokenizer,
    PreTrainedModel,
    BertForSequenceClassification,
    XLNetForSequenceClassification,
    AlbertForSequenceClassification,
)

from tqdm import tqdm

from adaptnlp.model import AdaptiveModel


logger = logging.getLogger(__name__)


class TransformersTokenTagger(AdaptiveModel):
    """Adaptive model for Transformer's Token Tagger Model

    Usage:
    ```python
    >>> tagger = TransformersTokenTagger.load("dbmdz/bert-large-cased-finetuned-conll03-english")
    >>> tagger.predict(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **tokenizer** - A tokenizer object from Huggingface's transformers (TODO)and tokenizers
    * **model** - A transformers token tagger model
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, model: PreTrainedModel):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        self.model = model

        # Setup cuda and automatic allocation of model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading and constructing this tagger

        * **model_name_or_path** - A key string of one of Transformer's pre-trained Token Tagger Model
        """
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForTokenClassification.from_pretrained(model_name_or_path)
        tagger = cls(tokenizer, model)
        return tagger

    def predict(
        self,
        text: Union[List[str], str],
        mini_batch_size: int = 32,
        grouped_entities: bool = True,
        **kwargs,
    ) -> List[List[Dict]]:
        """Predict method for running inference using the pre-trained token tagger model.
        Returns a list of lists of tagged entities.

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **mini_batch_size** - Mini batch size
        * **grouped_entities** - Set True to get whole entity span strings (Default True)
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Transformers tagger
        """
        if isinstance(text, str):
            text = [text]
        results: List[Dict] = []

        with torch.no_grad():
            # Tokenize and get dataset
            dataset = self._tokenize(text)
            dataloader = DataLoader(dataset, batch_size=mini_batch_size)

            logger.info(f"Running prediction on {len(dataset)} text sequences")
            logger.info(f"Batch size = {mini_batch_size}")
            for batch in tqdm(dataloader, desc="Predicting text"):
                self.model.eval()
                batch = tuple(t.to(self.device) for t in batch)

                if len(batch) == 3:
                    inputs = {
                        "input_ids": batch[0],
                        "attention_mask": batch[1],
                        "token_type_ids": batch[2],
                    }
                else:
                    inputs = {"input_ids": batch[0], "attention_mask": batch[1]}
                outputs = self.model(**inputs)

                # Iterate through batch for tagged token predictions
                for idx, pred in enumerate(outputs[0]):
                    entities = pred.cpu().detach().numpy()
                    input_ids = inputs["input_ids"].cpu().numpy()[idx]
                    tagged_entities = self._generate_tagged_entities(
                        entities=entities,
                        input_ids=input_ids,
                        grouped_entities=grouped_entities,
                    )
                    results += tagged_entities

        return results

    def _tokenize(
        self, sentences: Union[List[Sentence], Sentence, List[str], str]
    ) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with them """

        tokenized_text = self.tokenizer.batch_encode_plus(
            sentences,
            return_tensors="pt",
            pad_to_max_length=True,
        )

        # Bart, XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use token_type_ids
        if isinstance(
            self.model,
            (
                BertForSequenceClassification,
                XLNetForSequenceClassification,
                AlbertForSequenceClassification,
            ),
        ):
            dataset = TensorDataset(
                tokenized_text["input_ids"],
                tokenized_text["attention_mask"],
                tokenized_text["token_type_ids"],
            )
        else:
            dataset = TensorDataset(
                tokenized_text["input_ids"], tokenized_text["attention_mask"]
            )

        return dataset

    # `_group_entites` and `_generate_tagged_entities` modified from pipeline code snippet from Transformers
    def _group_entities(
        self, entities: List[dict], idx_start: int, idx_end: int
    ) -> Dict:
        """Returns grouped entities"""
        # Get the last entity in the entity group
        entity = entities[-1]["entity"]
        scores = np.mean([entity["score"] for entity in entities])
        tokens = [entity["word"] for entity in entities]

        entity_group = {
            "entity_group": entity,
            "score": np.mean(scores),
            "word": self.tokenizer.convert_tokens_to_string(tokens),
            "offsets": (idx_start, idx_end),
        }
        return entity_group

    def _generate_tagged_entities(
        self, entities: np.ndarray, input_ids: np.ndarray, grouped_entities: bool = True
    ) -> List[Dict]:
        """Generate full list of entities given tagged token predictions and input_ids"""

        score = np.exp(entities) / np.exp(entities).sum(-1, keepdims=True)
        labels_idx = score.argmax(axis=-1)

        answers = []
        entities = []
        entity_groups = []
        entity_group_disagg = []
        # Filter to labels not in `self.ignore_labels`
        filtered_labels_idx = [
            (idx, label_idx)
            for idx, label_idx in enumerate(labels_idx)
            if self.model.config.id2label[label_idx] not in ["O"]
        ]

        for idx, label_idx in filtered_labels_idx:
            # print(tokenizer.convert_ids_to_tokens(int(input_ids[idx])))
            entity = {
                "word": self.tokenizer.convert_ids_to_tokens(int(input_ids[idx])),
                "score": score[idx][label_idx].item(),
                "entity": self.model.config.id2label[label_idx],
                "index": idx,
            }
            last_idx, _ = filtered_labels_idx[-1]
            if grouped_entities:
                if not entity_group_disagg:
                    entity_group_disagg += [entity]
                    if idx == last_idx:
                        entity_groups += [
                            self._group_entities(
                                entity_group_disagg, idx - len(entity_group_disagg), idx
                            )
                        ]
                    continue

                # If the current entity is similar and adjacent to the previous entity, append it to the disaggregated entity group
                if (
                    entity["entity"] == entity_group_disagg[-1]["entity"]
                    and entity["index"] == entity_group_disagg[-1]["index"] + 1
                ):
                    entity_group_disagg += [entity]
                    # Group the entities at the last entity
                    if idx == last_idx:
                        entity_groups += [
                            self._group_entities(
                                entity_group_disagg, idx - len(entity_group_disagg), idx
                            )
                        ]
                # If the current entity is different from the previous entity, aggregate the disaggregated entity group
                else:
                    entity_groups += [
                        self._group_entities(
                            entity_group_disagg,
                            entity_group_disagg[-1]["index"] - len(entity_group_disagg),
                            entity_group_disagg[-1]["index"],
                        )
                    ]
                    entity_group_disagg = [entity]

            entities += [entity]

        # Append
        if grouped_entities:
            answers += [entity_groups]
        else:
            answers += [entities]

        return answers

    def train(
        self,
    ):
        raise NotImplementedError

    def evaluate(
        self,
    ):
        raise NotImplementedError


class FlairTokenTagger(AdaptiveModel):
    """Adaptive Model for Flair's Token Tagger...very basic

    Usage:
    ```python
    >>> tagger = FlairTokenTagger.load("en-sentiment")
    >>> tagger.predict(text="Example text", mini_batch_size=32)
    ```

    **Parameters:**

    * **model_name_or_path** - A key string of one of Flair's pre-trained Token tagger Model
    """

    def __init__(self, model_name_or_path: str):
        self.tagger = SequenceTagger.load(model_name_or_path)

    @classmethod
    def load(cls, model_name_or_path: str) -> AdaptiveModel:
        """Class method for loading a constructing this tagger

        * **model_name_or_path** - A key string of one of Flair's pre-trained Token tagger Model
        """
        tagger = cls(model_name_or_path)
        return tagger

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Predict method for running inference using the pre-trained token tagger model

        * **text** - String, list of strings, sentences, or list of sentences to run inference on
        * **mini_batch_size** - Mini batch size
        * **&ast;&ast;kwargs**(Optional) - Optional arguments for the Flair tagger
        """

        if isinstance(text, (Sentence, str)):
            text = [text]
        if isinstance(text[0], str):
            text = [Sentence(s) for s in text]
        self.tagger.predict(
            sentences=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )
        return text

    def train(
        self,
    ):
        raise NotImplementedError

    def evaluate(
        self,
    ):

        raise NotImplementedError


class EasyTokenTagger:
    """Token level classification models

    Usage:

    ```python
    >>> tagger = adaptnlp.EasyTokenTagger()
    >>> tagger.tag_text(text="text you want to tag", model_name_or_path="ner-ontonotes")
    ```
    """

    def __init__(self):
        self.token_taggers: Dict[AdaptiveModel] = defaultdict(bool)

    def tag_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        model_name_or_path: str = "ner-ontonotes",
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Tags tokens with labels the token classification models have been trained on

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        * **model_name_or_path** - The hosted model name key or model path
        * **mini_batch_size** - The mini batch size for running inference
        * **&ast;&ast;kwargs** - Keyword arguments for Flair's `SequenceTagger.predict()` method
        **return** - A list of Flair's `Sentence`'s
        """
        # Load Sequence Tagger Model and Pytorch Module into tagger dict
        if not self.token_taggers[model_name_or_path]:
            """
            self.token_taggers[model_name_or_path] = SequenceTagger.load(
                model_name_or_path
            )
            """
            # TODO: Find an alternative model-check method like an `is_available(model_name_or_path)
            # Check whether this is a Transformers or Flair Sequence Classifier model we're loading
            try:
                self.token_taggers[model_name_or_path] = FlairTokenTagger.load(
                    model_name_or_path
                )
            except FileNotFoundError:
                logger.info(
                    f"{model_name_or_path} not a valid Flair pre-trained model...checking transformers repo"
                )
                try:
                    self.token_taggers[
                        model_name_or_path
                    ] = TransformersTokenTagger.load(model_name_or_path)
                except ValueError:
                    logger.info("Not a valid model_name_or_path param")
                    return [Sentence("")]

        tagger = self.token_taggers[model_name_or_path]
        return tagger.predict(
            text=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )

    def tag_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """Tags tokens with all labels from all token classification models

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        * **mini_batch_size** - The mini batch size for running inference
        * **&ast;&ast;kwargs** - Keyword arguments for Flair's `SequenceTagger.predict()` method
        **return** A list of Flair's `Sentence`'s
        """
        if len(self.token_taggers) == 0:
            print("No token classification models loaded...")
            return Sentence()
        sentences = text
        for tagger_name in self.token_taggers.keys():
            sentences = self.tag_text(
                sentences,
                model_name_or_path=tagger_name,
                mini_batch_size=mini_batch_size,
                **kwargs,
            )
        return sentences



from typing import Union, List
from pathlib import Path
from abc import ABC, abstractmethod

from flair.data import Sentence


class AdaptiveModel(ABC):
    @abstractmethod
    def load(
        self,
        model_name_or_path: Union[str, Path],
    ):
        """ Load model into the `AdaptiveModel` object as alternative constructor """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        mini_batch_size: int = 32,
        **kwargs,
    ) -> List[Sentence]:
        """ Run inference on the model """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def train(
        self,
    ) -> None:
        """ Evaluate on the model """
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def evaluate(
        self,
    ) -> None:
        """ Evaluate on the model """
        raise NotImplementedError("Please Implement this method")



#!/usr/bin/env python
from pathlib import Path
from setuptools import setup, find_packages

version_file = Path(__file__).parent.joinpath("adaptnlp", "VERSION.txt")
version = version_file.read_text(encoding="UTF-8").strip()

with open("requirements.txt") as reqs_file:
    install_requires = reqs_file.read().splitlines()

with open("requirements_dev.txt") as reqs_dev_file:
    dev_requires = reqs_dev_file.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="adaptnlp",
    version=version,
    author="Andrew Chang",
    author_email="achang@novetta.com",
    packages=find_packages(),
    keywords=[
        "NLP",
        "flair",
        "Natural Language Processing",
        "Machine Learning",
        "ML",
        "torch",
        "pytorch",
        "NER",
    ],
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires
    },
    license="Apache 2.0",
    description="AdaptNLP: A Natural Language Processing Library and Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    zip_safe=True,
)



import logging
from typing import List, Dict, Union
from collections import defaultdict

from flair.data import Sentence
from flair.embeddings import (
    Embeddings,
    WordEmbeddings,
    StackedEmbeddings,
    FlairEmbeddings,
    DocumentPoolEmbeddings,
    DocumentRNNEmbeddings,
    TransformerWordEmbeddings,
)

logger = logging.getLogger(__name__)

FLAIR_PRETRAINED_MODEL_NAMES = {
    "multi-forward",
    "multi-backward",
    "multi-v0-forward",
    "multi-v0-backward",
    "multi-v0-forward-fast",
    "multi-v0-backward-fast",
    "en-forward",
    "en-backward",
    "en-forward-fast",
    "en-backward-fast",
    "news-forward",
    "news-backward",
    "news-forward-fast",
    "news-backward-fast",
    "mix-forward",
    "mix-backward",
    "ar-forward",
    "ar-backward",
    "bg-forward-fast",
    "bg-backward-fast",
    "bg-forward",
    "bg-backward",
    "cs-forward",
    "cs-backward",
    "cs-v0-forward",
    "cs-v0-backward",
    "da-forward",
    "da-backward",
    "de-forward",
    "de-backward",
    "de-historic-ha-forward",
    "de-historic-ha-backward",
    "de-historic-wz-forward",
    "de-historic-wz-backward",
    "es-forward",
    "es-backward",
    "es-forward-fast",
    "es-backward-fast",
    "eu-forward",
    "eu-backward",
    "eu-v1-forward",
    "eu-v1-backward",
    "eu-v0-forward",
    "eu-v0-backward",
    "fa-forward",
    "fa-backward",
    "fi-forward",
    "fi-backward",
    "fr-forward",
    "fr-backward",
    "he-forward",
    "he-backward",
    "hi-forward",
    "hi-backward",
    "hr-forward",
    "hr-backward",
    "id-forward",
    "id-backward",
    "it-forward",
    "it-backward",
    "ja-forward",
    "ja-backward",
    "nl-forward",
    "nl-backward",
    "nl-v0-forward",
    "nl-v0-backward",
    "no-forward",
    "no-backward",
    "pl-forward",
    "pl-backward",
    "pl-opus-forward",
    "pl-opus-backward",
    "pt-forward",
    "pt-backward",
    "pubmed-forward",
    "pubmed-backward",
    "sl-forward",
    "sl-backward",
    "sl-v0-forward",
    "sl-v0-backward",
    "sv-forward",
    "sv-backward",
    "sv-v0-forward",
    "sv-v0-backward",
    "ta-forward",
    "ta-backward",
}


class EasyWordEmbeddings:
    """Word embeddings from the latest language models

    Usage:

    ```python
    >>> embeddings = adaptnlp.EasyWordEmbeddings()
    >>> embeddings.embed_text("text you want embeddings for", model_name_or_path="bert-base-cased")
    ```
    """

    def __init__(self):
        self.models: Dict[Embeddings] = defaultdict(bool)

    def embed_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        model_name_or_path: str = "bert-base-cased",
    ) -> List[Sentence]:
        """Produces embeddings for text

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        * **model_name_or_path** - The hosted model name key or model path
        **return** - A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = Sentence(text)
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        else:
            sentences = text

        # Load correct Embeddings module
        if not self.models[model_name_or_path]:
            if (
                "flair" in model_name_or_path
                or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES
            ):
                self.models[model_name_or_path] = FlairEmbeddings(model_name_or_path)
            else:
                try:
                    self.models[model_name_or_path] = WordEmbeddings(model_name_or_path)
                except ValueError:
                    try:
                        self.models[model_name_or_path] = TransformerWordEmbeddings(
                            model_name_or_path
                        )
                    except ValueError:
                        raise ValueError(
                            f"Embeddings not found for the model key: {model_name_or_path}, check documentation or custom model path to verify specified model"
                        )
                        return Sentence("")
        embedding = self.models[model_name_or_path]
        return embedding.embed(sentences)

    def embed_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
        *model_names_or_paths: str,
    ) -> List[Sentence]:
        """Embeds text with all embedding models loaded

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        * **model_names_or_paths** -  A variable input of model names or paths to embed
        **return** - A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = Sentence(text)
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        else:
            sentences = text

        if model_names_or_paths:
            for embedding_name in model_names_or_paths:
                sentences = self.embed_text(
                    sentences, model_name_or_path=embedding_name
                )
        else:
            for embedding_name in self.models.keys():
                sentences = self.embed_text(
                    sentences, model_name_or_path=embedding_name
                )
        return sentences


class EasyStackedEmbeddings:
    """Word Embeddings that have been concatenated and "stacked" as specified by flair

    Usage:

    ```python
    >>> embeddings = adaptnlp.EasyStackedEmbeddings("bert-base-cased", "gpt2", "xlnet-base-cased")
    ```

    **Parameters:**

    * **&ast;embeddings** - Non-keyword variable number of strings specifying the embeddings you want to stack
    """

    def __init__(self, *embeddings: str):
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            if (
                "flair" in model_name_or_path
                or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES
            ):
                self.embedding_stack.append(FlairEmbeddings(model_name_or_path))
            else:
                try:
                    self.embedding_stack.append(WordEmbeddings(model_name_or_path))
                except ValueError:
                    try:
                        self.embedding_stack.append(
                            TransformerWordEmbeddings(model_name_or_path)
                        )
                    except ValueError:
                        raise ValueError(
                            f"Embeddings not found for the model key: {model_name_or_path}, check documentation or custom model path to verify specified model"
                        )

        assert len(self.embedding_stack) != 0
        self.stacked_embeddings = StackedEmbeddings(embeddings=self.embedding_stack)

    def embed_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """Stacked embeddings

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        **return** A list of Flair's `Sentence`s
        """
        # Convert into sentences
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text

        # Unlike flair embeddings modules, stacked embeddings do not return a list of sentences
        self.stacked_embeddings.embed(sentences)
        return sentences


class EasyDocumentEmbeddings:
    """Document Embeddings generated by pool and rnn methods applied to the word embeddings of text

    Usage:

    ```python
    >>> embeddings = adaptnlp.EasyDocumentEmbeddings("bert-base-cased", "xlnet-base-cased", methods["rnn"])
    ```

    **Parameters:**

    * **&ast;embeddings** - Non-keyword variable number of strings referring to model names or paths
    * **methods** - A list of strings to specify which document embeddings to use i.e. ["rnn", "pool"] (avoids unncessary loading of models if only using one)
    * **configs** - A dictionary of configurations for flair's rnn and pool document embeddings
    ```python
    >>>example_configs = {"pool_configs": {"fine_tune_mode": "linear", "pooling": "mean", },
    ...                   "rnn_configs": {"hidden_size": 512,
    ...                                   "rnn_layers": 1,
    ...                                   "reproject_words": True,
    ...                                   "reproject_words_dimension": 256,
    ...                                   "bidirectional": False,
    ...                                   "dropout": 0.5,
    ...                                   "word_dropout": 0.0,
    ...                                   "locked_dropout": 0.0,
    ...                                   "rnn_type": "GRU",
    ...                                   "fine_tune": True, },
    ...                  }
    ```
    """

    __allowed_methods = ["rnn", "pool"]
    __allowed_configs = ("pool_configs", "rnn_configs")

    def __init__(
        self,
        *embeddings: str,
        methods: List[str] = ["rnn", "pool"],
        configs: Dict = {
            "pool_configs": {"fine_tune_mode": "linear", "pooling": "mean"},
            "rnn_configs": {
                "hidden_size": 512,
                "rnn_layers": 1,
                "reproject_words": True,
                "reproject_words_dimension": 256,
                "bidirectional": False,
                "dropout": 0.5,
                "word_dropout": 0.0,
                "locked_dropout": 0.0,
                "rnn_type": "GRU",
                "fine_tune": True,
            },
        },
    ):
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Check methods
        for m in methods:
            assert m in self.__class__.__allowed_methods

        # Set configs for pooling and rnn parameters
        for k, v in configs.items():
            assert k in self.__class__.__allowed_configs
            setattr(self, k, v)

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            if (
                "flair" in model_name_or_path
                or model_name_or_path in FLAIR_PRETRAINED_MODEL_NAMES
            ):
                self.embedding_stack.append(FlairEmbeddings(model_name_or_path))
            else:
                try:
                    self.embedding_stack.append(WordEmbeddings(model_name_or_path))
                except ValueError:
                    try:
                        self.embedding_stack.append(
                            TransformerWordEmbeddings(model_name_or_path)
                        )
                    except ValueError:
                        raise ValueError(
                            f"Embeddings not found for the model key: {model_name_or_path}, check documentation or custom model path to verify specified model"
                        )

        assert len(self.embedding_stack) != 0
        if "pool" in methods:
            self.pool_embeddings = DocumentPoolEmbeddings(
                self.embedding_stack, **self.pool_configs
            )
            print("Pooled embedding loaded")
        if "rnn" in methods:
            self.rnn_embeddings = DocumentRNNEmbeddings(
                self.embedding_stack, **self.rnn_configs
            )
            print("RNN embeddings loaded")

    def embed_pool(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """Stacked embeddings


        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        **return** - A list of Flair's `Sentence`s
        """
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text
        self.pool_embeddings.embed(sentences)
        return sentences

    def embed_rnn(
        self,
        text: Union[List[Sentence], Sentence, List[str], str],
    ) -> List[Sentence]:
        """Stacked embeddings

        * **text** - Text input, it can be a string or any of Flair's `Sentence` input formats
        **return** - A list of Flair's `Sentence`s
        """
        if isinstance(text, str):
            sentences = [Sentence(text)]
        elif isinstance(text, list) and all(isinstance(t, str) for t in text):
            sentences = [Sentence(t) for t in text]
        elif isinstance(text, Sentence):
            sentences = [text]
        else:
            sentences = text
        self.rnn_embeddings.embed(sentences)
        return sentences
